/*
   hst.c -- PIKT history processing routines

   $Header: /var/cvs/PIKT/pikt/src/piktd/hst.c,v 1.46 2007/03/12 10:02:35 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2007 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
*/

#include "../lib/common.h"
#include "piktd.h"

FILE *hstin;
FILE *hstout;

char *hsttmpfile;

hist *histcurr;

char *keystr = (char *) NULL;
int keystrlen;
int keyseplen;

/* do a lookup of an alarm, returning a pointer to its data structure
 */
alrm *
lookupalarm (char *name)
{
	alrm *alp;

	foreach (alp, alarms)
		if (! strcmp(alp->name, name))
			return alp;
	return (alrm *) NULL;
}

/* given an input log statement, verify (a) that the log file exists and
   (b) the name of this file matches between what's in the history file and
   what's in the current alarm script
 */
void
verifylog (char *logname)
{
	static char fnnam[] = "verifylog";
	struct stat fbuf;

	if (stat(logname, &fbuf) != 0) {
		logreturn(PIKTLOG, LOG_WARNING, ERRID_FILE_STAT_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, stat() error for %s", ALMNAM, logname);
		HSTERR = TRUE;
	} else if (strcmp(INPUTLOGNAME,logname)) {
		logreturn(PIKTLOG, LOG_WARNING, ERRID_HISTORY_LOG_FILE_NAME_MISMATCH, __FILE__, __LINE__, fnnam, FALSE, "in %s, log name in .hst file (%s) doesn't match log name in %s file (%s)", ALMNAM, logname, isalert ? ".alt" : "script", INPUTLOGNAME);
		HSTERR = TRUE;
	}
	return;
}

/* verify that the current key var shadows a legitimate var and is in
   the proper sequence
 */
void
verifykey (int vrndx, signed long *arndx, char *arkey)
{
	static char fnnam[] = "verifykey";

	/* if this vrndx not found in the new history list, or if this
	   vrndx is not next in the key sequence, signify error, else
	   advance to the next key
	 */
	if (! NEWHST(vrndx)) {
		logreturn(PIKTLOG, LOG_WARNING, ERRID_HISTORY_FILE_KEY_INDEX_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "in %s, key index %d in .hst file doesn't match any current history var index", ALMNAM, vrndx);
		HSTERR = TRUE;
	} else if (keycurr && (keycurr->vrndx != vrndx)) {
		logreturn(PIKTLOG, LOG_WARNING, ERRID_HISTORY_KEY_INDEX_OUT_OF_SEQUENCE, __FILE__, __LINE__, fnnam, FALSE, "in %s, key index %d out of sequence", ALMNAM, vrndx);
		HSTERR = TRUE;
	} else if (keycurr) {
		keycurr = keycurr->next;
	}
	return;
}

/* verify that the current var has not changed, from the history file to
   the present alarm, in various ways that would invalidate use of its
   history vals
 */
void
verifyvar (int vrndx, char *name)
{
	static char fnnam[] = "verifyvar";
	int vx;

	/* if this var name has changed slots, or is a reserved name,
	   or is not a current var, or if this var name has
	   dropped off the new history list, or if this vrndx has
	   dropped off the new history list, or if this varndx was
	   found already in the old history list, signify error
	 */
	if (((vx = varndx(name)) == NIL) || ((vx != vrndx) && (VALTYP(vx) & FIXEDVAL))) {
		logreturn(PIKTLOG, LOG_WARNING, ERRID_HISTORY_FILE_VAR_INDEX_CHANGED, __FILE__, __LINE__, fnnam, FALSE, "in %s, var %s changed index, else reserved var or fn name, else not found", ALMNAM, name);
		HSTERR = TRUE;
	} else if (! NEWHST(vrndx)) {
		logreturn(PIKTLOG, LOG_WARNING, ERRID_HISTORY_FILE_VAR_INDEX_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "in %s, var index %d in .hst file missing from new history list", ALMNAM, vrndx);
		HSTERR = TRUE;
	} else if (OLDHST(vrndx)) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_HISTORY_FILE_VAR_INDEX_DUPLICATED, __FILE__, __LINE__, fnnam, FALSE, "in %s, duplicate var index %d in history", ALMNAM, vrndx);
		HSTERR = TRUE;
	} else
		OLDHST(vrndx) = TRUE;
	return;
}

/*
void
verifyfld (int vrndx, signed long *arndx, char *arkey, int fld)
{
	static char fnnam[] = "verifyfld";
	spc *sp;
	int fldnum;

	/ * if the var's fldnum has changed, signify error * /
	sp = lookupspc(vrndx, arndx, arkey);
	if (sp) {
		fldnum = sp->fldnum;
		if (fldnum != fld) {
			if (VARNAM(vrndx))
				logreturn(PIKTLOG, LOG_WARNING, ERRID_HISTORY_FILE_VAR_FLDNUM_MISMATCH, __FILE__, __LINE__, fnnam, FALSE, "in %s, current var fldnum (%d) and history var fldnum (%d) mismatch for var %s", ALMNAM, fldnum, fld, VARNAM(vrndx));
			else
				logreturn(PIKTLOG, LOG_WARNING, ERRID_HISTORY_FILE_VAR_FLDNUM_MISMATCH, __FILE__, __LINE__, fnnam, FALSE, "in %s, current var fldnum (%d) and history var fldnum (%d) mismatch for varndx %d", ALMNAM, fldnum, fld, vrndx);
			HSTERR = TRUE;
		}
	} else {
		if (VARNAM(vrndx))
			logreturn(PIKTLOG, LOG_WARNING, ERRID_ARRAY_INDEX_NULL, __FILE__, __LINE__, fnnam, FALSE, "in %s, for var %s, null array index", ALMNAM, VARNAM(vrndx));
		else
			logreturn(PIKTLOG, LOG_WARNING, ERRID_ARRAY_INDEX_NULL, __FILE__, __LINE__, fnnam, FALSE, "in %s, for vrndx %d, null array index", ALMNAM, vrndx);
		HSTERR = TRUE;
	}
	return;
}
*/

/* verify that the current var's arydim has not changed; if it has, this
   invalidates use of its history vals
 */
void
verifydim (int vrndx, int dim)
{
	static char fnnam[] = "verifydim";

	/* if the var's arydim has changed, signify error */
	if (ARYDIM(vrndx) != dim) {
		if (VARNAM(vrndx))
			logreturn(PIKTLOG, LOG_WARNING, ERRID_HISTORY_FILE_VAR_ARYDIM_MISMATCH, __FILE__, __LINE__, fnnam, FALSE, "in %s, current var arydim (%d) and history var arydim (%d) mismatch for var %s", ALMNAM, ARYDIM(vrndx), dim, VARNAM(vrndx));
		else
			logreturn(PIKTLOG, LOG_WARNING, ERRID_HISTORY_FILE_VAR_ARYDIM_MISMATCH, __FILE__, __LINE__, fnnam, FALSE, "in %s, current var arydim (%d) and history var arydim (%d) mismatch for vrndx %d", ALMNAM, ARYDIM(vrndx), dim, vrndx);
		HSTERR = TRUE;
	}
	return;
}

/* verify that the current var's dattype has not changed; if it has, this
   invalidates use of its history vals
 */
void
verifytyp (int vrndx, int typ)
{
	static char fnnam[] = "verifytyp";

	/* if the var's dattype has changed, signify error */
	if (DATTYP(vrndx) != typ) {
		if (VARNAM(vrndx))
			logreturn(PIKTLOG, LOG_WARNING, ERRID_HISTORY_FILE_VAR_DATTYP_MISMATCH, __FILE__, __LINE__, fnnam, FALSE, "in %s, current var dattype (%s) and history var dattype (%s) mismatch for var %s", ALMNAM, DATTYP(vrndx) ? "NUMBER" : "STRING", typ ? "NUMBER" : "STRING", VARNAM(vrndx));
		else
			logreturn(PIKTLOG, LOG_WARNING, ERRID_HISTORY_FILE_VAR_DATTYP_MISMATCH, __FILE__, __LINE__, fnnam, FALSE, "in %s, current var dattype and history var dattype mismatch for vrndx %d", ALMNAM, vrndx);
		HSTERR = TRUE;
	}
	return;
}

/* verify that the current history val corresponds to either an old or a
   new history var
 */
void
verifyval (int vrndx, signed long *arndx, char *arkey)
{
	static char fnnam[] = "verifyval";

	/* if this varndx has dropped off the new history list, or if this
	   varndx is missing from the old history list, signify error */
	if (! NEWHST(vrndx)) {
		logreturn(PIKTLOG, LOG_WARNING, ERRID_HISTORY_FILE_OLD_VAL_INDEX_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "in %s, val index %d in .hst file missing from new history list", ALMNAM, vrndx);
		HSTERR = TRUE;
	} else if (! OLDHST(vrndx)) {
		logreturn(PIKTLOG, LOG_WARNING, ERRID_HISTORY_FILE_NEW_VAL_INDEX_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "in %s, val index %d in .hst file missing from old history list", ALMNAM, vrndx);
		HSTERR = TRUE;
	}
	return;
}

/* open the history file; read in and parse the history file; make history
   lookup keys
 */
void
openhistory (char *hstfile)
{
	static char fnnam[] = "openhistory";
	struct stat fbuf;

	if (stat(hstfile, &fbuf) != 0) {
		/* always true when alert run first time
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_STAT_ERROR, __FILE__, __LINE__, fnnam, TRUE, "stat() error for %s", hstfile);
		 */
		;
	} else {
		if (fbuf.st_size > 0) {
			if ((hstin = fopen(hstfile, "r")) == NULL) {
				logexit(PIKTLOG, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fopen() error for %s", hstfile);
			} else {
				scanpass = 1;
				hstlex();
				if (fclose(hstin)) {
					logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fclose() error for %s", hstfile);
				}
			}
			if ((hstin = fopen(hstfile, "r")) == NULL) {
				logexit(PIKTLOG, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fopen() error for %s", hstfile);
			} else {
				lineno = 1;
				alarmcurr = (alrm *) NULL;
				keycurr = (ky *) NULL;
				scanpass = 2;
				hstparse();
				if (fclose(hstin)) {
					logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fclose() error for %s", hstfile);
				}
			}
		}
	}
	hsttmpfile = (char *) mymalloc((strlen(hstdir)+strlen(altcurr)+6) * sizeof(char), fnnam);
	sprintf(hsttmpfile, "%s/%s.tmp", hstdir, altcurr);
	if ((hstout = fopen(hsttmpfile, "w")) == NULL)
		logexit(PIKTLOG, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fopen() error for %s", hsttmpfile);
	return;
}

/* close new history file; replace old history file with new, revised
   history file
 */
void
closehistory (char *hstfile)
{
	static char fnnam[] = "closehistory";

	if (fclose(hstout))
		logexit(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fclose() error for %s", hsttmpfile);
	if (remove(hstfile))
		/* always true when alert run first time
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_REMOVE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "remove() error for %s", hstfile);
		 */
		;
	if (rename(hsttmpfile, hstfile))
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_RENAME_ERROR, __FILE__, __LINE__, fnnam, TRUE, "rename() error for %s and %s", hsttmpfile, hstfile);
	chmod(hstfile, (mode_t)(S_IRUSR|S_IWUSR));	/* 600 */
	free(hsttmpfile);
	return;
}

/* write alarm name and key info (if any) to new history file, if called for
 */
int
beginhistory (void)
{
	static char fnnam[] = "beginhistory";
	register int i, k;
	ky *kyp;

	/* write alarm name to history file */
	if (HSTTYP /* || KEYS || INPUTLOGNAME */)
		fprintf(hstout, "A %s\n", ALMNAM);
	if (HSTTYP & (FIXEDVAL|FREEVAL|BEGINVAL|ENDVAL) /* KEYS */) {
		/* write alarm var parameters to history file */
		for (i=1; i<=VRNDXHI; i++) {
			if (NEWHST(i)) {
				if (VARNAM(i))
					fprintf(hstout, "R %d %d %d %d %s\n", i, VALTYP(i), ARYDIM(i), DATTYP(i), VARNAM(i));
				else {
					logreturn(PIKTLOG, LOG_ERR, ERRID_VAR_NAME_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "in %s, missing name for var %d", ALMNAM, i);
					STATUS = ERROR;
					return ERROR;
				}
			}
		}
		/* write alarm keys parameters to history file */
		if (KEYS) {
			foreach (kyp, KEYS)
				if (kyp->arkey) {
					fprintf(hstout, "K %d %c%s%c\n", kyp->vrndx, KEYBEGIN, kyp->arkey, KEYEND);
				} else {
					fprintf(hstout, "K %d", kyp->vrndx);
					for (k=0; k<=ARYDIM(kyp->vrndx); k++)
						fprintf(hstout, " %ld", kyp->arndx[k]);
					fprintf(hstout, "\n");
				}
		}
	}
	return 0;
}

/* write global vals to new history file, if called for; record log file last
   position in new history file, if called for
 */
int
endhistory (void)
{
	vl *vlp, *initvlp;
	ky *kyp;
	register int i, j, k;

	for (i=1; i<=VRNDXHI; i++) {
		if (NEWHST(i) && (VALTYP(i) & (BEGINVAL|FREEVAL|ENDVAL))) {
			if ((ARYTYP(i) == STRING) && ARKEYS(i)) {
				foreach (kyp, ARKEYS(i)) {
					vlp = lookupvl(valnow, i, arndxnul, kyp->arkey, STRING);
					/* val->num ? no, just write str */
					if (vlp && vlp->str) {
						if (vlp->key) {
							fprintf(hstout, "V %d %c%s %s%c %s\n", i, KEYBEGIN, KEYSEP, vlp->key, KEYEND, vlp->str);
						} else {
							fprintf(hstout, "V %d %c%s", i, KEYBEGIN, KEYSEP);
							for (k=0; k<=ARYDIM(i); k++)
								fprintf(hstout, " %ld", vlp->ndx[k]);
							fprintf(hstout, "%c %s\n", KEYEND, vlp->str);
						}
					}
				}
			} else {
				for (j=0; j<=9; j++) {
					vlp = initvlp = valnow->vals[i][j];
					while (vlp) {
						/* vlp->num ? no, just write str */
						if (vlp->str) {
							if (vlp->key) {
								fprintf(hstout, "V %d %c%s %s%c %s\n", i, KEYBEGIN, KEYSEP, vlp->key, KEYEND, vlp->str);
							} else {
								fprintf(hstout, "V %d %c%s", i, KEYBEGIN, KEYSEP);
								for (k=0; k<=ARYDIM(i); k++)
									fprintf(hstout, " %ld", vlp->ndx[k]);
								fprintf(hstout, "%c %s\n", KEYEND, vlp->str);
							}
						}
						vlp = vlp->next;
						if (vlp == initvlp)
							break;
					}
				}
			}
		}
	}
	if ((INPUTTYP == LOGINPUT) && INPUTSRC)
		fprintf(hstout, "L %lu %lu %lu %s\n", (unsigned long)inputino, (unsigned long)inputsize, inputposend, INPUTSRC);
	return 0;
}

/* free vals in the val struct, either when disposing of a history val
   that's no longer needed, or when assigning new valnows
 */
void
freevals (val **valp, int type)
{
	register int i, j;
	vl *vlp, *initvlp, *tvlp;

	for (i=1; i<=VRNDXHI; i++) {
		if (VALTYP(i) & type) {
			for (j=0; j<=9; j++) {
				vlp = initvlp = (*valp)->vals[i][j];
				while ((tvlp = vlp)) {
					FREE(vlp->key);
					FREE(vlp->str);
					vlp = vlp->next;
					free(tvlp);
					if (vlp == initvlp)
						break;
				}
				(*valp)->vals[i][j] = (vl *) NULL;
			}
		}
	}
	/* if we are not freeing valnow or valprc, we are freeing history
           values for the current alarm, so let's dispose of its val struct
	 */
	if ((*valp != valnow) && (*valp != valprc)) {
		FREEVOID(*valp);
	}
	return;
}

/* build suitable key string for current input line
 */
int
makekeystr (void)
{
	static char fnnam[] = "makekeystr";
	vl *vlp;
	ky *kyp;
	char *keystrtmp;

	if (KEYS /* && ! HSTERR */) {	/* even if HSTERR, make keystr, else
					   valnows might not be written in
					   recordvalnows()
					 */
		foreach (kyp, KEYS) {
			if ((vlp = lookupvl(valnow, kyp->vrndx, kyp->arndx, kyp->arkey, ARYTYP(kyp->vrndx)))) {
				keystrtmp = (char *) mymalloc((keystrlen+(keystrlen ? 1 : 0)+strlen(vlp->str)+1)*sizeof(char), fnnam);
				sprintf(keystrtmp, "%s%s%s", (keystrlen ? keystr : nullstring), (keystrlen ? KEYSEP : nullstring), vlp->str);
				FREE(keystr);
				keystr = keystrtmp;
				keystrlen = strlen(keystr);
			} else {
				logreturn(PIKTLOG, LOG_ERR, ERRID_VAR_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "in %s, lookupvl() failure for %s", ALMNAM, VARNAM(kyp->vrndx));
				return ERROR;
			}
		}
	}
	return 0;
}

/* write valnows for current input line to new history file, if called for
 */
int
recordvalnows (void)
{
	vl *vlp, *initvlp;
	ky *kyp;
	register int i, j, k;

	if (KEYS /* && ! rulcurr */) {
		for (i=1; i<=VRNDXHI; i++) {
			if (NEWHST(i) && (VALTYP(i) & (FIXEDVAL|FREEVAL))) {
				if ((ARYTYP(i) == STRING) && ARKEYS(i)) {
					foreach (kyp, ARKEYS(i)) {
						vlp = lookupvl(valnow, i, arndxnul, kyp->arkey, STRING);
						/* val->num ? no, just write str */
						if (vlp && vlp->str && keystr) {
							if (vlp->key) {
								fprintf(hstout, "V %d %c%s %s%c %s\n", i, KEYBEGIN, keystr, vlp->key, KEYEND, vlp->str);
							} else {
								fprintf(hstout, "V %d %c%s", i, KEYBEGIN, keystr);
								for (k=0; k<=ARYDIM(i); k++)
									fprintf(hstout, " %ld", vlp->ndx[k]);
								fprintf(hstout, "%c %s\n", KEYEND, vlp->str);
							}
						}
					}
				} else {
					for (j=0; j<=9; j++) {
						vlp = initvlp = valnow->vals[i][j];
						while (vlp) {
							/* val->num ? no, just write str */
							if (vlp->str && keystr) {
								if (vlp->key) {
									fprintf(hstout, "V %d %c%s %s%c %s\n", i, KEYBEGIN, keystr, vlp->key, KEYEND, vlp->str);
								} else {
									fprintf(hstout, "V %d %c%s", i, KEYBEGIN, keystr);
									for (k=0; k<=ARYDIM(i); k++)
										fprintf(hstout, " %ld", vlp->ndx[k]);
									fprintf(hstout, "%c %s\n", KEYEND, vlp->str);
								}
							}
							vlp = vlp->next;
							if (vlp == initvlp)
								break;
						}
					}
				}
			}
		}
	}
	return 0;
}

/* register a history line
 */
void
reghistory (char *lin)
{
	static char fnnam[] = "reghistory";
        hist *nhp;

	nhp = (hist *) mymalloc(sizeof(hist), fnnam);
	nhp->next = (hist *) NULL;
	nhp->line = lin;
	if (! HISTORY) {
		HISTORY = nhp;
	} else {
		histcurr->next = nhp;
	}
	histcurr = nhp;
	return;
}

/* free all history lines for the current script
 */
void
freehistory (hist **histp)
{
	hist *hp, *thp;

	hp = *histp;
	while ((thp = hp)) {
		FREE(hp->line);
		hp = hp->next;
		free(thp);
	}
	*histp = (hist *) NULL;
	return;
}

/* in quit or error situations, recall previous script history and write
   to the current history file
 */
void
recallhistory (void)
{
	hist *hp;

	hp = HISTORY;
	while (hp) {
		fprintf(hstout, "%s", hp->line);
		hp = hp->next;
	}
	return;
}
